use std::{pin::Pin, sync::Arc};

use axum::{
    extract::{Path, State},
    http::StatusCode,
    response::IntoResponse,
};
use futures::TryStreamExt;
use tokio::sync::Mutex;
use tokio_stream::Stream;
use tokio_util::bytes::Bytes;

use crate::{crypt, state::AppState};

pub async fn handle_secure(
    Path(data): Path<String>,
    state: State<Arc<Mutex<AppState>>>,
) -> Result<impl IntoResponse, MyAnyhow> {
    let app_state = state.lock().await;
    let decrypt = crypt::decryt(data, app_state.secret_key.clone().unwrap())?;
    std::mem::drop(app_state);

    let bytes = handle(decrypt, state).await?;
    Ok(axum::body::Body::from_stream(bytes))
}

pub async fn handle_unsecure(
    Path(src): Path<String>,
    state: State<Arc<Mutex<AppState>>>,
) -> Result<impl IntoResponse, MyAnyhow> {
    let bytes = handle(src, state).await?;

    Ok(axum::body::Body::from_stream(bytes))
}

async fn handle(
    src: String,
    state: State<Arc<Mutex<AppState>>>,
) -> anyhow::Result<impl Stream<Item = Result<Bytes, anyhow::Error>> + Send> {
    let mut app_state = state.lock().await;

    match app_state.storage_pool.retrieve(src.clone()).await {
        Some(stream) => Ok(stream),
        None => {
            let save_stream = fetch(src.clone()).await?;
            app_state
                .storage_pool
                .save(src.clone(), save_stream)
                .await?;

            Ok(app_state.storage_pool.retrieve(src).await.unwrap())
        }
    }
}

async fn fetch(
    src: String,
) -> anyhow::Result<Pin<Box<dyn Stream<Item = Result<Bytes, anyhow::Error>> + Send>>> {
    let client = reqwest::Client::new();
    let res = client.get(src).send().await?;
    let stream = res.bytes_stream().map_err(|err| anyhow::anyhow!(err));

    Ok(Box::pin(stream))
}

pub struct MyAnyhow(anyhow::Error);

impl IntoResponse for MyAnyhow {
    fn into_response(self) -> axum::response::Response {
        (StatusCode::INTERNAL_SERVER_ERROR, self.0.to_string()).into_response()
    }
}

impl<T> From<T> for MyAnyhow
where
    T: Into<anyhow::Error>,
{
    fn from(value: T) -> Self {
        Self(value.into())
    }
}
